/*
 *   bit_utils is a support module which provides bit oriented utilities
 *   Copyright (C) 2013 Elio Blanca <eblanca76@users.sourceforge.net>
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

/* this will enable 64 bit integers as file offset on 32 bit posix machines */
#define _FILE_OFFSET_BITS 64

#ifdef __MINGW32__
#define __USE_MINGW_FSEEK         /* Request mingw internal implementation of fseeko64 */
#define __USE_MINGW_ANSI_STDIO 1  /* Select a more ANSI C99 compatible implementation of printf() and friends. */
#endif /* __MINGW32__ */

#include <stdio.h>

#ifdef __MINGW32__
#define off_t     off64_t
#define fseeko    fseeko64
#define ftello    ftello64
#endif /* __MINGW32__ */

#include "bit_utils.h"

PART23BUF p23b[NP23B]; 
int p23b_in=0,p23b_out=0;
int p23b_pos=0;  /* offset of the current bit inside the current output block (p23b_out) */

void initBits(FILEBITS *fb,FILE *fp, off_t eobs)
{
    int i;
    fb->fp = fp;
    fb->nLastBits = 0;
    fb->location = 8*ftello(fp);
    fb->latest_bit = eobs*8;  /* end_of_bitstream is already known, it's located at id3pos */
    for (i=0; i<NP23B; i++)
    {
        p23b[i].pos = 0;
        p23b[i].len = 0;
    }
}

off_t ftellBits(FILEBITS *fb)
{
    return fb->location;
}

void fseekBits(FILEBITS *fb,off_t offset,int whence)
{
    switch(whence)
    {
        case SEEK_SET:
            fb->location = offset;
            break;
        case SEEK_CUR:
            fb->location += offset;
            break;
        case SEEK_END:
            fb->location = fb->latest_bit;
            if (offset<0) fb->location+=offset; /* cannot move beyond EObitstream */
            break;
    }
    fseeko(fb->fp,fb->location/8,SEEK_SET);
/*
 * I choose not to check for EOF here, since fseeking at EOF doesn't set eof() on `true'.
 * Next calls to getc will, and subsequent checks for eof() will be able to detect the failure.
 */

    /* Ok, now fill the register lastBits and prepare for next readBits calls */
    fb->lastBits = (unsigned char)getc(fb->fp);
    fb->lastBits = fb->lastBits<<8;
    fb->lastBits |= (unsigned char)getc(fb->fp);
    fb->lastBits = fb->lastBits<<8;
    fb->lastBits |= (unsigned char)getc(fb->fp);
    fb->lastBits = fb->lastBits<<8;
    fb->lastBits |= (unsigned char)getc(fb->fp);

    fb->nLastBits = 32 - (int)(fb->location % 8);
}

unsigned long int readBits(FILEBITS *fb,int n)
/*
 * This routine provides a bit-oriented reading facility.
 * It stores bytes read from the stream into a buffer called lastBits
 * (it is managed as a scrolling register, the output is from the most significant bits)
 * and counts the valid (unused) ones into nLastBits (the remaining least significant bits).
 * User calls cause the high order bits to be extracted and returned as a result,
 * discarding them from the nLastBits count.
 * So, when the left valid bits are <= than 24, then this routine can fullfill the register
 * with (at least) a new byte.
 * Note: even if an EOF check is performed, no flag is returned about the result.
 */
{
    unsigned long int ret = 0;
    int k,tmp;
    unsigned char c;

    fb->location += n;

/*
 * I think this first `while' cycle is here since in the early versions of fseekBits
 * the lastBits buffer was not filled after a seek operation, so this check was mandatory
 * As a safe move, I will keep it here for a while and see whether the alert is printed or not
 */
    while(fb->nLastBits <= 24)
    {
printf("*** while statement @ %d ***\n",__LINE__);
        tmp = getc(fb->fp);
        c = tmp;
        fb->lastBits = fb->lastBits << 8;
        if (tmp != EOF) fb->lastBits = fb->lastBits | c;
        fb->nLastBits += 8;
    }
    /* first of all, prepare the bits I already have */
    if (n <= fb->nLastBits)
        k = n;
    else
        k = fb->nLastBits;

    /* put into `ret' the bits I need from lastBits field and decrease nLastBits */
    ret = (fb->lastBits >> (fb->nLastBits - k)) & ((1 << k)-1);
    fb->nLastBits -= k;
    /* here `n' is updated; if still n>0 then I'll need some more bits */
    n -= k;

    /* fullfill the bit buffer */
    while(fb->nLastBits <= 24)
    {
        tmp = getc(fb->fp);
        c = tmp;
        fb->lastBits = fb->lastBits << 8;
        if (tmp != EOF)
            fb->lastBits = fb->lastBits | c;
        fb->nLastBits += 8;
    }

    if (n)
    {
        ret = ret << n;
        ret = ret | ((fb->lastBits >> (fb->nLastBits - n)) & ((1 << n)-1));
        fb->nLastBits -= n;
    }
    return ret;
}

/*
 * `seek_p23b' searches for the requested bit inside the p23b structure
 * (given `pos' as an absolute position) and seeks to that bit when it's found,
 * returning 1 as OK (the indexes `p23b_out' and `p23b_pos' are updated as well).
 * When not found, the function returns zero.
 */
char seek_p23b(off_t pos)
{
//printf("pos=%lld\n",pos);
    int i;
    char ret=0;
    for(i=0; i<NP23B; i++)
    {
        if (p23b[i].pos <= pos && pos < p23b[i].pos+(off_t)p23b[i].len-(p23b[i].pos & 7))
        {
            /* `p23b_out' will point the entry containing the requested bit */
            p23b_out = i;
            /* `p23b_pos' will be the relative position of the requested bit inside the selected entry */
            /* Note: I am sure (p23b[*].pos & 7) is always zero */
            p23b_pos = (int)(pos-p23b[p23b_out].pos+(p23b[p23b_out].pos & 7));
            ret=1;
            break;
        }
    }
//if (!ret) printf("seek_p23b : pos=%lld, i=%d, p23b[i].pos=%lld, p23b[i].len=%d\n",pos,i,p23b[i].pos,p23b[i].len);
    return ret;
}

/*
 * This function seeks forward `len' bits into the p23b structure
 * and sets `p23b_out' and `p23b_pos' indexes accordingly.
 * If unable to seek (not enough data) it returns the spare bits,
 * else zero.
 */
int skip_p23b(int len)
{
    while(len)
    {
        /*printf("skip_p23b : len=%d, p23b_in=%d, p23b_out=%d, p23b_pos=%d, p23b[p23b_out].pos=%d, p23b[p23b_out].len=%d\n",
            len,p23b_in,p23b_out,p23b_pos,p23b[p23b_out].pos,p23b[p23b_out].len);*/
        if (p23b[p23b_out].len-p23b_pos < len)
        {
            len -= p23b[p23b_out].len-p23b_pos;
            p23b_out = (p23b_out+1) % NP23B;
            p23b_pos = p23b[p23b_out].pos & 7;
            if (p23b_in == p23b_out)
                break;
        }
        else
        {
            p23b_pos += len;
            len = 0;
        }
    }
    return len;
}

/*
 * This function seeks back `len' bits into the p23b structure
 * and sets `p23b_out' and `p23b_pos' indexes accordingly.
 * If unable to seek (not enough data) it returns the spare bits,
 * else zero.
 */
int skipback_p23b(int len)
{
    while(len)
    {
        /*printf("skipback_p23b : len=%d, p23b_in=%d, p23b_out=%d, p23b_pos=%d, p23b[p23b_out].pos=%d, p23b[p23b_out].len=%d\n",
            len,p23b_in,p23b_out,p23b_pos,p23b[p23b_out].pos,p23b[p23b_out].len);*/
        if (p23b[p23b_out].len-(p23b[p23b_out].pos & 7) < len)
        {
            len -= p23b[p23b_out].len-(p23b[p23b_out].pos & 7);

            p23b_out--;
            if (p23b_out == -1) p23b_out = NP23B-1;

            p23b_pos = p23b[p23b_out].len-1;
            if (p23b_out == p23b_in)
                break;
        }
        else
        {
            p23b_pos -= len;
            len = 0;
        }
    }
    return len;
}

off_t tell_p23b(void)
{
//printf("p23b_pos=%u, p23b_out=%u, p23b[p23b_out].pos=%u, (p23b[p23b_out].pos & 7)=%u\n",p23b_pos,p23b_out,p23b[p23b_out].pos,(p23b[p23b_out].pos & 7));
    return p23b_pos + p23b[p23b_out].pos - (p23b[p23b_out].pos & 7);
}

void load_into_p23b(FILEBITS *fb,int len)
{
/*
 * `load_into_p23b' is called after reading the whole side information,
 * which is either 9, 17 or 32 bytes long. Maybe there is protection by CRC-16
 * following the mpeg header (32 bit). So I can be assured ftellBits returns a
 * value which is multiple of 8, and ftellBits()&7 will always be zero.
 * What's the point in always taking this value into account?
 */
    int block_len=len/8;

    /* store the offset of the beginning of part2 data (bits) with respect to the entire file stream */
    p23b[p23b_in].pos = ftellBits(fb);
    /* length of the data block */
    p23b[p23b_in].len = len + (int)(p23b[p23b_in].pos & 7);
//if (len&7) printf("`load_into_p23b': len=%d\n",len);
//if (p23b[p23b_in].pos&7) printf("`load_into_p23b': pos=%lld\n",(long long)p23b[p23b_in].pos);
/*
 * I have to use `fseek' here instead of `fseekBits' since `fseekBits'
 * has to be used in conjunction with `readBits', not `fread'.
 */
    fseeko(fb->fp,ftellBits(fb)/8,SEEK_SET);
    if (block_len>LARGEST_BUFFER)
    {
        printf("Warning: fread requested for a block too large - fixing.\n");
        block_len=LARGEST_BUFFER;
    }
    if (fread(p23b[p23b_in].buf,1,block_len,fb->fp)!=(unsigned)block_len)
        printf("Warning: fread failure in `load_into_p23b'.\n"); /* this should never happen! */
    /* this seek will take me to the first bit of the next frame's syncword */
    fseekBits(fb,p23b[p23b_in].pos-(p23b[p23b_in].pos&7)+(off_t)p23b[p23b_in].len,SEEK_SET);
    if (p23b_in == p23b_out)
        p23b_pos = p23b[p23b_in].pos & 7;
    p23b_in = (p23b_in+1) % NP23B;
}

// read byte-aligned 8 bits
unsigned char readOneByte_p23b(void)
{
    unsigned char ret = 0;
//printf("p23b_in=%d, p23b_out=%d, p23b_pos=%d(%c)\n",p23b_in,p23b_out,p23b_pos,(p23b_pos&7)?'*':' ');
//    if (p23b_in == p23b_out) return 0;
    if (p23b_in != p23b_out)
    {
        for(;;)
        {
            p23b_pos = (p23b_pos+7) & ~7; /* is this some kind of increment? -- YES! it works like ceil(actual byte) */
            if (p23b[p23b_out].len-p23b_pos >= 8)
            {
                ret = p23b[p23b_out].buf[p23b_pos>>3];
                p23b_pos += 8;
                break;
            }
            else
            {
                p23b_out = (p23b_out+1) % NP23B;
                p23b_pos = p23b[p23b_out].pos & 7;
                if (p23b_out == p23b_in)
                    break;
            }
        }
    }
    return ret;
}

