/*
 *   mp3guessenc is an mp3 encoder-guesser
 *   Copyright (C) 2002-2010 Naoki Shibata
 *   Copyright (C) 2011-2013 Elio Blanca <eblanca76@users.sourceforge.net>
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public
 *   License along with this library; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

/* mp3guessenc version 0.2 by Naoki Shibata */
/* This program is still under development. */


#define VER_PKG_NAME   "mp3guessenc"
#define VER_MAJ        0
#define VER_MIN        25
#define VER_PATCH_LEV  2
#define CODENAME       "List Games"

#define REL            2  /* release type 0:stable, 1:alpha, anything else means beta */

/*
 * release type is provided via RELEASE_STABLE/RELEASE_ALPHA/RELEASE_BETA macros
 *
 * So please, do not use REL for release check! 
 * Again, check which is true among RELEASE_STABLE/RELEASE_ALPHA/RELEASE_BETA
 */

#ifdef REL
#define RELEASE_STABLE (REL==0)
#define RELEASE_ALPHA  (REL==1)
#define RELEASE_BETA   ((REL!=0)&&(REL!=1))
#else
#define RELEASE_STABLE 0
#define RELEASE_ALPHA  1
#define RELEASE_BETA   0
#endif

/* this will enable 64 bit integers as file offset on 32 bit posix machines */
#define _FILE_OFFSET_BITS 64

#ifdef __MINGW32__
#define __USE_MINGW_FSEEK         /* Request mingw internal implementation of fseeko64 */
#define __USE_MINGW_ANSI_STDIO 1  /* Select a more ANSI C99 compatible implementation of printf() and friends. */
#endif /* __MINGW32__ */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#ifdef __MINGW32__
#define off_t     off64_t
#define fseeko    fseeko64
#define ftello    ftello64
#endif /* __MINGW32__ */

#include "mp3guessenc.h"
#include "bit_utils.h"
#include "decode.h"
#include "scrambled.h"
#define CMD_CODENAME_LENGTH 80 /* max length of a string provided via command line */

int samples_tab[2][3] = {{ 384,1152,1152},
                         { 384,1152, 576}};

int samplerate_tab [3][4] = {{11025,99999,22050,44100},
                             {12000,99999,24000,48000},
                             { 8000,99999,16000,32000}};

int bitrate_tab[2][3][16] = {
  { {128, 32, 64, 96,128,160,192,224,256,288,320,352,384,416,448,},
    {128, 32, 48, 56, 64, 80, 96,112,128,160,192,224,256,320,384,},
    {128, 32, 40, 48, 56, 64, 80, 96,112,128,160,192,224,256,320,} },

  { {128, 32, 48, 56, 64, 80, 96,112,128,144,160,176,192,224,256,},
    {128,  8, 16, 24, 32, 40, 48, 56, 64, 80, 96,112,128,144,160,},
    {128,  8, 16, 24, 32, 40, 48, 56, 64, 80, 96,112,128,144,160,} }
};

char sideinfo_tab[2][2] = {{32,17},{17,9}};

//////////////////////////////////////////////
// Mpeg header utilities

char head_check(int head, char relaxed)
/*
 * Integrity checks for an mpeg header (4 bytes)
 * Return values:
 *  0: invalid header
 *  1: valid header
 * when `relaxed' is nonzero, then the check on bitrate index is skipped
 * but only for layerIII !
 */
{
    unsigned char ly=(head&HEADER_FIELD_LAYER)>>17;            /* layer index */
    unsigned char bi=(head&HEADER_FIELD_BITRATE)>>12;          /* bitrate index */
    unsigned char mp=(head&HEADER_FIELD_MPEG_ID)>>19;          /* mpeg version */
    unsigned char md;                                          /* audio mode */

    if ((head&HEADER_FIELD_SYNC)!=HEADER_FIELD_SYNC) return 0; /* check for sync bits */
    if (mp == 1) return 0;                                     /* check for reserved value in mpeg version field */
    if (!ly) return 0;                                         /* check for reserved value in layer field */
    if (!relaxed || ly&2)
        if (bi == 15) return 0;                                    /* check for invalid bitrate index */
    if (((head&HEADER_FIELD_SAMPRATE)>>10) == 3 ) return 0;    /* check for reserved value in sample rate field */
    if ((head&HEADER_FIELD_EMPHASIS) == 2) return 0;           /* check for reserved value in emphasis field */

    /* the header is valid and the mpeg frame is supported! */

    if ((ly != 2)||(mp != 3)) return 1;
    /* if mpeg 1 layerII, then check for not allowed bitrate-mode combinations */
    if (bi == 0) return 1; /* free format bitstreams allow for any bitrate-mode combinations */
    md=(head&HEADER_FIELD_CHANNELS)>>6;
    if (bi<4 && md!=3) return 0;
    if (bi==5 && md!=3) return 0;
    if (bi>10 && md==3) return 0;
    /* all went ok, this mpeg1layerII header is valid */
    return 1;
}

int get_framesize(int head)
/*
 * "Just give me an audio mpeg header,
 * I will give you in return its size"
 * Works with mpeg1, mpeg2, mpeg2.5, all layers
 */
{
    unsigned char lsf,layer,bitrate_index,srate_index,mpeg,padding,samplesfactor=1;
    int size;

    lsf           = 1-((head&HEADER_FIELD_LSF)>>19);
    layer         = 3-((head&HEADER_FIELD_LAYER)>>17);   /* this goes from 0 (layerI) up to 2 (layerIII) */
    bitrate_index = (head&HEADER_FIELD_BITRATE)>>12;
    srate_index   = (head&HEADER_FIELD_SAMPRATE)>>10;
    mpeg          = (head&HEADER_FIELD_MPEG_ID)>>19;
    padding       = (head&HEADER_FIELD_PADDING)>>9;

/*
 * This `samplesfactor' is used to compensate for rounding errors when managing integers only. The
 * size formula is right just without this (aside from the pad*4) but it keeps not returning
 * multiple-of-4 values. So this way it is fixed for working in any case with layers II and III.
 * For layer I, I just multiply by 4 before returning the size.
 */
    if (!layer) samplesfactor=4;

    size = samples_tab[lsf][layer]/samplesfactor*bitrate_tab[lsf][layer][bitrate_index]*125/samplerate_tab[srate_index][mpeg]+(int)padding;

    if (layer) return size;
    else return size<<2;
}

///////////////////////////////////////////////////////

off_t resync_mp3(FILE *fp, int *h, off_t pos, int *scanned, char relaxed)
/*
 * Scan the stream for sync bits
 * return: the offset of the next frame found
 *         and the full mpeg header (stored in `h', 32 bits)
 *         -1 in case of invalid pos / no valid header found
 * Further, into `scanned' we can read how many bytes
 * have been read before exiting.
 * You can give an offset into `pos', or
 * the current file position can be used setting
 * pos=RESYNC_POS_CURRENT.
 * relaxed=1 makes the scan take as valid headers also
 * those with 1111b in the bitrate field
 * (old versions of lame used to set this in vbr tags
 * of free format streams)
 * Note: if a valid mpeg frame is found, then
 * before exiting a `fseek' is done in order to
 * restore the file pointer to the frame start
 * Note 2: despite the name, this function seeks any
 * audio mpeg frame
 */
{
    int head=0;

    if (pos==RESYNC_POS_CURRENT)
        pos=ftello(fp);
    else
    {
        if (pos<0) return -1; /* fseeking to negative offset is invalid */
        fseeko(fp,pos,SEEK_SET);
    }

    head  = (unsigned char)getc(fp); head = head << 8;
    head |= (unsigned char)getc(fp); head = head << 8;
    head |= (unsigned char)getc(fp); head = head << 8;
    for(*scanned=0;*scanned<SCAN_SIZE;(*scanned)++)
    {
        head |= (unsigned char)getc(fp);
        if (feof(fp)) return -1;
        if (head_check(head,relaxed))
        {
            *h=head;
            fseeko(fp,pos+(off_t)(*scanned),SEEK_SET);
            return pos+(off_t)(*scanned);
        }
        head = head << 8;
    }

    return -1;
}

/*
 * this is used to set up `regular' bitrates only
 * in case of free format bitstreams, the setup has already
 * been done into main, thanks to GetInfoTag routine
 */
void setup_framesize(int head, streamInfo *si, detectionInfo *di)
{
    int i, slot, headbase;

    if (di->lay==1) slot=4;
    else slot=1;

    headbase=head&HEADER_ANY_BUT_BITRATE_AND_PADDING_FIELDS;
    for (i=1;i<15;i++)
    {
        si->unpadded[i].frameSize=get_framesize(headbase|(i<<12));
        si->padded[i].frameSize=si->unpadded[i].frameSize+slot;
    }
}

char freeformat(int head)
{
    return ((head&HEADER_FIELD_BITRATE)==0);
}

char layerIII(int head)
{
    return ((head&HEADER_FIELD_LAYER)==0x20000);
}

void scan_ancillary(detectionInfo *di, streamInfo *si, int ancill_Len, char *buf)
{
    if (ancill_Len==0)
    {   /* no ancillary data into the current frame */
        /* this is not expected to happen both in mp3pro and mp3surround
           streams, hence I reset the `enhanced features' bytes */
        di->enhSignature[0] = 0;
        di->enhSignature[1] = 0;
        return;
    }

    /* check for an OFL block into the very first frame */
    if (si->totFrameNum==1 && di->lay==3 /* first frame in a layerIII stream */
        &&
        (
        (di->id!=3 && ancill_Len>20 && (unsigned char)buf[0]==0xE0 && (unsigned char)buf[1]==0xBA)
        ||
        (di->id==3 && ancill_Len> 9 && (unsigned char)buf[0]==0xB4 && ((unsigned char)buf[1]==0x08 || (unsigned char)buf[1]==0x04))
        )
       )
    {
        di->ofl=1;
        if (di->id!=3)
        {   /* this is an mp3pro stream with OFL */
            /* at least 20 bytes of ancillary data into the very first frame are used for length informations */
            si->encDelay=ExtractI2(buf+14);
            si->orig_samples=ExtractI4(buf+16);
            di->enhSignature[0] = di->enhSignature[0] & buf[0];
            di->enhSignature[1] = di->enhSignature[1] & buf[1];
        }
        else
        {
            /* not an mp3pro stream */
            /* 10 bytes of ancillary data into the very first frame are used for length informations */
            si->encDelay=ExtractI2(buf+1);
            si->orig_samples=ExtractI4(buf+3);
            if ((unsigned char)buf[1]==0x08)
            {
                /* mp3Surround with OFL */
                si->encDelay+=576;
                if (ancill_Len>10) di->enhSignature[0] = di->enhSignature[0] & buf[10];
                if (ancill_Len>11) di->enhSignature[1] = di->enhSignature[1] & buf[11];
            }
        }
    }
    else
    {
        if (ancill_Len>0) di->enhSignature[0] = di->enhSignature[0] & buf[0];
        if (ancill_Len>1) di->enhSignature[1] = di->enhSignature[1] & buf[1];
    }
    di->ancillaryData += ancill_Len;
    if (ancill_Len>si->ancill_max) si->ancill_max = ancill_Len;
    if (ancill_Len<si->ancill_min) si->ancill_min = ancill_Len;
    extract_enc_string(di->encoder_string,buf,ancill_Len);
}

void scan_layerIII(streamInfo *si, FILEBITS *fb, detectionInfo *di, int head, off_t startOfFrame, int framesize, off_t *endOfLastPart3, char *buf)
{
    int main_data_begin,part2_3_length,part1_length,gg=-1;
    off_t pos_main_data_begin;
    char gr,ngr=((head&HEADER_FIELD_LSF)>>19)+1,ch,nch=(((head&HEADER_FIELD_CHANNELS)>>6)==3)?1:2,scfsi_band,window_switching_flag,block_type,mixed_block_flag;

    /* (si->id == 3) means mpeg version 1 */
    if (di->id==3)
        main_data_begin = readBits(fb,9);
    else
        main_data_begin = readBits(fb,8);

    if (main_data_begin > si->reservoirMax)
        si->reservoirMax = main_data_begin;

    if (di->id==3)
    {   /* si->mode == 3 means single channel mode */
        if (si->mode==3)
            readBits(fb,5); // private_bits
        else
            readBits(fb,3); // private_bits
    }
    else
    {
        if (si->mode==3)
            readBits(fb,1); // private_bits
        else
            readBits(fb,2); // private_bits
    }

    if (di->id == 3)
    {
        for(ch=0;ch<nch;ch++)
            for(scfsi_band=0;scfsi_band<4;scfsi_band++)
                di->usesScfsi = readBits(fb,1) || di->usesScfsi;
    }

    part2_3_length = 0;

    for(gr=0;gr<ngr;gr++)
    {
        for(ch=0;ch<nch;ch++)
        {
            part2_3_length += readBits(fb,12);
            readBits(fb,9); // big_values
            gg = readBits(fb,8); // global_gain
//            readBits(fb,9+8); /* a single call brings time saving */
            if (gg>si->max_global_gain[(int)ch]) si->max_global_gain[(int)ch]=(unsigned char)gg;
            if (gg<si->min_global_gain[(int)ch]) si->min_global_gain[(int)ch]=(unsigned char)gg;
            if (di->id == 3)
                readBits(fb,4); // scalefac_compress
            else
                readBits(fb,9); // scalefac_compress
            window_switching_flag = readBits(fb,1);
            if (window_switching_flag == 1)
            {
                block_type = readBits(fb,2); // block_type
                mixed_block_flag = readBits(fb,1); // mixed_block_flag
//                for(region=0;region<2;region++) readBits(&fb,5); // table_select
//                for(window=0;window<3;window++) readBits(&fb,3); // subblock_gain
                readBits(fb,5+5+3+3+3); /* a single call brings time saving */
                /* This block detection needs to be fixed, it doesn't match with lame results */
                /* NOW FIXED! */
                if (block_type == 2)
                {
                    if (mixed_block_flag)
                        di->blockCount[2]++;  /* this granule contains mixed blocks */
                    else
                        di->blockCount[1]++;  /* this granule contains short blocks */
                }
                else
                    di->blockCount[3]++;      /* this granule contains switch blocks */
            }
            else
            {
//                int region;
//                for(region=0;region<3;region++) readBits(&fb,5); // table_select
//                readBits(&fb,4); // region0_count
//                readBits(&fb,3); // region1_count
                readBits(fb,5+5+5+4+3); /* a single call brings time saving */
                di->blockCount[0]++;      /* this granule contains long blocks */
            }
            if (di->id == 3)
                readBits(fb,1); // preflag
            di->usesScalefacScale = readBits(fb,1) || di->usesScalefacScale; // scalefac_scale
            readBits(fb,1); // count1table_select
        }
    }
    /* we've just finished reading the side information */
    part1_length = (int)(ftellBits(fb) - startOfFrame); /* startOfFrame actually points to the sync bits of the current frame header */
    /* part1_length may have one of the values 72+32, 136+32 or 256+32 (bit) (mpeg header is taken into account, add 16 bit in case of crc16 protection) */
    // load part 2 and part 3 into buffer

//printf("framesize=%d, part1_length=%d, load : %d bit from %lld\n",framesize,part1_length,framesize*8-part1_length,ftellBits(fb));
    load_into_p23b(fb,framesize*8-part1_length);

    /* first of all, I need to properly set `pos_main_data_begin',
       that is, the offset of the first bit of part2+part3 data of
       the current frame --
       in case of insufficient data, `pos_main_data_begin' is set to -1 */
    if (main_data_begin == 0)
    {
        pos_main_data_begin = startOfFrame+part1_length;
    }
    else
        /* if main_data_begin != 0 then it has to be meant as a negative number => it's a real backpointer */
        if (seek_p23b(startOfFrame-1)) /* is there any data before the actual frame ? */
        {
            if (skipback_p23b(main_data_begin*8-1)) /* if so, then skip back */
            {
                /* not enough data collected */
                printf("frame %d : More bits in reservoir are needed to decode this frame.\n",si->totFrameNum);
                pos_main_data_begin = -1;
            }
            else
            {
                /* the seek operation returned ok */
                pos_main_data_begin = tell_p23b();
//printf("endOfLastPart3=%d,mdb=%d\n",endOfLastPart3,pos_main_data_begin);
            }
        }
        else
        {
            /* no data before current part1 data */
            printf("frame %d : more bits in reservoir are needed to decode this frame.\n",si->totFrameNum);
            pos_main_data_begin = -1;
        }

    if (pos_main_data_begin != -1) /* was the beginning of main_data found? */
    {   // decode part2, part3 and ancillary bits -- NOT YET!
        if (*endOfLastPart3 != -1) /* was I able to seek the end of part2+part3 data for the previous frame? */
        {
            if (seek_p23b(*endOfLastPart3))
            {
                char *p;
                unsigned char c;
                seek_p23b((*endOfLastPart3+7)&~7); /* seek at the end of part3 data from the previous frame */
//                if (pos_main_data_begin-tell_p23b() >= 8)
//                {
                p = buf;
                while(pos_main_data_begin-tell_p23b() >= 8)
                {
                    c = readOneByte_p23b();
                    if ((p-buf)<LARGEST_BUFFER)
                        *p = c;
                    else
                        printf("%s @ %u: *** Stack smashing prevention! ***\nBuffer overflow avoided, double-check the code!!! (pointer=%u)\n",__FILE__,__LINE__,(unsigned int)(p-buf));
                    p++;
                }
                scan_ancillary(di,si,(int)(p-buf),buf);
//                }
            }
        }
    }
    *endOfLastPart3 = -1;

    /* this block skips part2+part3 data and sets endOfLastPart3 for the next iteration */
//printf("pos_main_data_begin=%d, part2_3_length=%d\n",pos_main_data_begin,part2_3_length);
    if (pos_main_data_begin != -1)
    {
        if (seek_p23b(pos_main_data_begin))
        {
            if (skip_p23b(part2_3_length) == 0)
                *endOfLastPart3 = tell_p23b();
            else
                printf("B!\n");
        }
        else
            printf("A!\n");
    }

    return;
}

void scan(FILE *input_file, off_t *pos, streamInfo *si, detectionInfo *di, off_t id3pos, char *buf)
{
    int head,i,framesize,head_pattern,mask;
    off_t endOfLastPart3=-1,oldpos;
    char last_frame_incomplete=0;
    FILEBITS fb;
/*
 * I'm sure I'm about to read the first valid audio frame of the stream,
 * since `scan' is called right after a successfull resync_mp3().
 * Now resync_mp3() is called again in order to get the value of `head'
 * and to properly set the file pointer.
 */
    oldpos=*pos=resync_mp3(input_file, &head, *pos, &i, 0);
/*
 * Gathering basic properties of the mpeg stream, from now on I will assume
 * the other frames have the same mpeg version/layer/freq/etc...
 */

    si->freeformat = (head&HEADER_FIELD_BITRATE)==0;
    si->lsf        = 1-((head&HEADER_FIELD_LSF)>>19);
    di->id         = (head&HEADER_FIELD_MPEG_ID)>>19;   /* 0=mpeg2.5, 1=resvd, 2=mpeg2, 3=mpeg1 */
    di->freq       = samplerate_tab[(head&HEADER_FIELD_SAMPRATE)>>10][di->id];
    di->lay        = 4-((head&HEADER_FIELD_LAYER)>>17); /* so it is 1=layerI, 2=layerII, 3=layerIII */
    si->mode       = (head&HEADER_FIELD_CHANNELS)>>6;   /* channel mode (stereo, Jstereo, Dchannel, mono) */

    si->crc        = (head&HEADER_FIELD_CRC) >> 16;
    si->copyright  = (head&HEADER_FIELD_COPYRIGHT) >> 3;
    si->original   = (head&HEADER_FIELD_ORIGINAL) >> 2;
    si->emphasis   = head&HEADER_FIELD_EMPHASIS;

    setup_framesize(head, si, di);

    for(i=0;i<15;i++)
    {
        si->bitrateCount[i]        = 0;
        si->unpadded[i].frameCount = 0;
        si->padded[i].frameCount   = 0;
    }
    for(i=0;i<5;i++) di->modeCount[i]  = 0;
    for(i=0;i<4;i++) di->blockCount[i] = 0;

    /* mask selection: looks like layer I/II when encoding jstereo, switch between different stereo
       modes (they put 0x0, 0x1 or 0x2 in the channels field), while layer III only uses
       jstereo frames (0x1 in the channels field) and puts further informations into the mode
       extension field */
    if (layerIII(head)||((head&HEADER_FIELD_CHANNELS>>6)==3))
      mask=HEADER_CONSTANT_FIELDS_STRICT;
    else
      mask=HEADER_CONSTANT_FIELDS;

    /* using a bit mask will let me check a whole header in a single `if' statement */
    head_pattern = head & mask;
    initBits(&fb,input_file,id3pos);

//    while (*pos!=-1 && *pos<id3pos)
    while (*pos!=-1)
    {
        if (*pos!=oldpos)
        {
            printf("Resync done at frame %d, skipped %d bytes.\n",si->totFrameNum,(int)(*pos-oldpos));
            endOfLastPart3=-1;
        }

        /* this call is needed in order to properly set the `fb' struct after resync */
        /* the file pointer is moved right after the header */
        fseekBits(&fb,8*(*pos)+32,SEEK_SET);

        /* now check the header we've just read */
        if ((head&mask)!=head_pattern)
        {
            si->nSyncError++;
            /* do I need a real full-featured error detection routine? mhhh... */
            printf("Sync error at frame %d, offset %lld. Seeking for resync...\n",si->totFrameNum,(long long)*pos);
            *pos=resync_mp3(input_file, &head, (*pos)+1, &i, 0);
            continue;
        }

        /* Let's go calculate the frame size */
        i=(head&HEADER_FIELD_BITRATE)>>12;

        if (head&HEADER_FIELD_PADDING)
            framesize=si->padded[i].frameSize;
        else
            framesize=si->unpadded[i].frameSize;

        /* check integrity for this frame */
        if (*pos+(off_t)framesize>id3pos)
        {
            /* cannot read the whole frame, so skip the rest */
            printf("Frame %d at offset %lld appears to be incomplete.\n %d bytes available, %d expected. Skipping...\n",si->totFrameNum,(long long)*pos,(int)(id3pos-(*pos)),framesize);
            last_frame_incomplete = 1;    /* this flag prevents from printing `unexpected data at...' later */
//            *pos=resync_mp3(input_file, &head, (*pos)+1, &i, 0);
            *pos=-1;
//            continue;
            break;
        }

        /* if this point is reached, then I'm sure the frame
           is readable as a whole, so I assume it's valid */
        if (!(head&HEADER_FIELD_CRC)) readBits(&fb,16);

        if (head&HEADER_FIELD_PADDING)
        {
            di->usesPadding = 1;
            si->padded[i].frameCount++;
        }
        else
            si->unpadded[i].frameCount++;
        si->bitrateCount[i]++;

        if (layerIII(head))
            scan_layerIII(si,&fb,di,head,(*pos)*8,framesize,&endOfLastPart3,buf);

        si->totFrameLen += (off_t)framesize;
        si->totFrameNum ++;

        if ((head&HEADER_FIELD_CHANNELS)>>6 == 1) /* if jstereo, then update mode_extension data */
        {
            di->modeCount[(head&HEADER_FIELD_MODEXT)>>4]++;
            /* when handling layer I/II we may find a stereo frame at the beginning, jstereo frames
               (signaling the file is encoded as jstereo) may be encountered later */
            si->mode=1;
        }
        else
            di->modeCount[4]++;

        *pos += (off_t)framesize;

        oldpos=*pos;
        *pos=resync_mp3(input_file, &head, *pos, &i, 0);
    } /* END OF `while' CYCLE */

    if (layerIII(head))
    {
        if (endOfLastPart3 != -1)
        {
            if (seek_p23b(endOfLastPart3))
            {
                // extract ancillary
                off_t end_of_last_frame;
                char *p;
                unsigned char c;
                seek_p23b(oldpos*8-1); /* replaced startOfFrame with oldpos */
                end_of_last_frame = tell_p23b()+1;
                seek_p23b((endOfLastPart3+7)&~7);
//                if (end_of_last_frame-tell_p23b() >= 8)
//                {
                p = buf;
                while(end_of_last_frame-tell_p23b() >= 8)
                {
                    c = readOneByte_p23b();
                    if ((p-buf)<LARGEST_BUFFER)
                        *p = c;
                    else
                        printf("%s @ %u: *** Stack smashing prevention! ***\nBuffer overflow avoided, double-check the code!!! (pointer=%u)\n",__FILE__,__LINE__,(unsigned int)(p-buf));
                    p++;
                }
                scan_ancillary(di,si,(int)(p-buf),buf);
//                }
            }
        }
    }
    /* `oldpos' now stores the last `to-be-checked' byte offset, is it exactly the same as id3pos? */
    if (oldpos<id3pos && !last_frame_incomplete)
        printf("Unexpected data at %lld (length in bytes: %d).\n",(long long)oldpos,(int)(id3pos-oldpos));
}
/////////////////////////////////////////////////////////////////
///////////////////      END SCAN      //////////////////////////
/////////////////////////////////////////////////////////////////

/* This routine prints lots of detailed informations about the processed mpeg bitstream */
void show_stream_info(streamInfo si, detectionInfo di, VBRTAGDATA *vbrTag)
{
    int i;
    char max_len,blk_max_len;
    double duration,sum;

    printf("\nDetected MPEG stream version %s layer ",!(di.id)? "2.5" : (di.id==2)? "2" : "1");
    for (i=0;i<di.lay;i++) printf("I");
    printf(", details follow.\n  File size                : %lld bytes\n",(long long)si.filesize);
    printf("  Audio stream size        : %lld bytes",(long long)si.totFrameLen);
    /* if the file has a header tag, then show the total size */
    if (vbrTag->infoTag!=TAG_NOTAG)
    {
        printf (" (including tag: %lld)",(long long)(si.totFrameLen+(off_t)vbrTag->frameSize));
        /* just a simple check */
        if (si.totFrameLen+(off_t)vbrTag->frameSize>si.filesize)
            printf(" (!!!)");
    }
    else
        if (si.totFrameLen>si.filesize)
            printf(" (!!!)");
    duration=(double)si.totFrameNum*(double)samples_tab[si.lsf][di.lay-1]/(double)di.freq;
    printf("\n  Length                   : %u:%02u:%02u.%03.0f",((int)duration/3600),(((int)duration%3600)/60),((int)duration%60),(1000*(duration-(int)duration)));
    if (duration>=60)
        printf(" (%1.3f seconds)",duration);
    printf("\n  Data rate                : ");
//    if (si.totFrameNum)
    printf("%1.1f kbps",(double)si.totFrameLen*(double)di.freq*0.008/((double)si.totFrameNum*(double)samples_tab[si.lsf][di.lay-1]));
//    else printf("0 kbps");
    if (si.freeformat) printf(" (free format bitstream)");
    printf("\n  Number of frames         : ");
    max_len=printf("%d",si.totFrameNum);
    if (di.lay==3)
        printf("\n  Blocks per frame         : %d (granules per frame %d, channels per granule %d)",((di.id==3)?2:1)*((si.mode==3)?1:2),(di.id==3)?2:1,(si.mode==3)?1:2);
    printf("\n  Audio samples per frame  : %d\n",samples_tab[si.lsf][di.lay-1]);
    printf("  Audio frequency          : %d Hz\n",di.freq);
    if (si.encDelay!=-1)
        printf("  Encoder delay            : %d samples\n",si.encDelay);
//    if (vbrTag->lametag[0])
    if (di.lameTag)
        si.orig_samples=si.totFrameNum*samples_tab[si.lsf][di.lay-1]-vbrTag->encDelay-vbrTag->encPadding;
    if (si.orig_samples!=-1)
        printf("  Length of original audio : %d samples\n",si.orig_samples);
    printf("  Encoding mode            : ");
    switch (si.mode)
    {
        case 3:
            printf("mono");
            break;
        case 2:
            printf("dual channel");
            break;
        case 1:
            printf("joint stereo");
            break;
        case 0:
            printf("stereo");
            break;
    }
    if (di.lay==3)
    {
        printf("\n  Min global gain          : ");
        if (si.mode==3) /* Mono? */
            printf("c=%3d",si.min_global_gain[0]);
        else
            printf("l=%3d  r=%3d",si.min_global_gain[0],si.min_global_gain[1]);
        printf("\n  Max global gain          : ");
        if (si.mode==3) /* Mono? */
            printf("c=%3d",si.max_global_gain[0]);
        else
            printf("l=%3d  r=%3d",si.max_global_gain[0],si.max_global_gain[1]);
    }
    printf("\nFlags\n  Error protection : %s\n",si.crc ? "no":"yes");
    printf("  Copyrighted      : %s\n",si.copyright ? "yes":"no");
    printf("  Original         : %s\n",si.original ? "yes":"no");
    printf("  Emphasis         : %s\n\n",si.emphasis == 0 ? "none" : si.emphasis == 1 ? "50/15ms" : "CCITT");

    if (si.mode==1) /* Jstereo? */
    {
        sum=(double)(di.modeCount[0]+di.modeCount[1]+di.modeCount[2]+di.modeCount[3]+di.modeCount[4]);
        if (di.lay==3)
        {
            printf("Mode extension: stereo mode frame count\n");
            if (di.modeCount[0]) printf("  Simple stereo                 : %*d (%4.1f%%)\n",max_len,di.modeCount[0],(double)di.modeCount[0]*100.0/sum);
            if (di.modeCount[1]) printf("  Intensity stereo              : %*d (%4.1f%%)\n",max_len,di.modeCount[1],(double)di.modeCount[1]*100.0/sum);
            if (di.modeCount[2]) printf("  Mid-side stereo               : %*d (%4.1f%%)\n",max_len,di.modeCount[2],(double)di.modeCount[2]*100.0/sum);
            if (di.modeCount[3]) printf("  Intensity and mid-side stereo : %*d (%4.1f%%)\n",max_len,di.modeCount[3],(double)di.modeCount[3]*100.0/sum);
            printf("  -----------------------------\n  sum                           ");
        }
        else
        {
            printf("Mode extension: intensity stereo applied in\n");
            if (di.modeCount[0]) printf("  Bands  4 to 31     : %*d (%4.1f%%)\n",max_len,di.modeCount[0],(double)di.modeCount[0]*100.0/sum);
            if (di.modeCount[1]) printf("  Bands  8 to 31     : %*d (%4.1f%%)\n",max_len,di.modeCount[1],(double)di.modeCount[1]*100.0/sum);
            if (di.modeCount[2]) printf("  Bands 12 to 31     : %*d (%4.1f%%)\n",max_len,di.modeCount[2],(double)di.modeCount[2]*100.0/sum);
            if (di.modeCount[3]) printf("  Bands 16 to 31     : %*d (%4.1f%%)\n",max_len,di.modeCount[3],(double)di.modeCount[3]*100.0/sum);
            if (di.modeCount[4]) printf("Simple stereo frames : %*d (%4.1f%%)\n",max_len,di.modeCount[4],(double)di.modeCount[4]*100.0/sum);
            printf("  ------------------\n  sum                ");
        }
        printf(": %d\n\n",(int)sum);
    }

    if (di.lay==3)
    {
        sum=(double)(di.blockCount[0]+di.blockCount[1]+di.blockCount[2]+di.blockCount[3]);
        /*
         * It is not hard to say that blk_max_len and max_len are related, since
         * granule number is related to frame number. Granule number may be the same
         * value as frame number if single channel and one granule per channel (mpeg2/2.5).
         * It may be frame number times two if either two channels and one granule per
         * channel or single channel and two granules per channel (mpeg1).
         * At last, it may be the frame number times four when two channels and two
         * granules per channel are used.
         * So blk_max_len may require at most a single digit more than max_len.
         */
        blk_max_len=max_len+1;
        printf("Block usage\n");
        if (di.blockCount[0]) printf("  Long block granules   : %*d (%4.1f%%)\n",blk_max_len,di.blockCount[0],(double)di.blockCount[0]*100.0/sum);
        if (di.blockCount[2]) printf("  Mixed block granules  : %*d (%4.1f%%)\n",blk_max_len,di.blockCount[2],(double)di.blockCount[2]*100.0/sum);
        if (di.blockCount[3]) printf("  Switch block granules : %*d (%4.1f%%)\n",blk_max_len,di.blockCount[3],(double)di.blockCount[3]*100.0/sum);
        if (di.blockCount[1]) printf("  Short block granules  : %*d (%4.1f%%)\n",blk_max_len,di.blockCount[1],(double)di.blockCount[1]*100.0/sum);
        printf("  ---------------------\n  sum                   : %*d\n\n",blk_max_len,(int)sum);

        printf("Ancillary data\n  Total amount   : %d bytes (%3.1f%%)\n  Bitrate        : %1.1f kbps\n",di.ancillaryData,(double)di.ancillaryData/(double)si.totFrameLen*100.0,(duration!=0)?(double)di.ancillaryData*0.008/duration:0);
        if (si.ancill_min!=LARGEST_BUFFER)
            printf("  Min packet     : %d bytes\n  Max packet     : %d bytes\n",si.ancill_min,si.ancill_max);
        printf("Max reservoir                       : %d bytes.\n",si.reservoirMax);
        printf("Scalefactor scaling used            : %s\n",(di.usesScalefacScale)?"yes":"no");
        printf("Scalefactor select information used : %s\n",(di.usesScfsi)?"yes":"no");
    }
    printf("Padding used       %s: %s\n",(di.lay==3)?"                 ":"",(di.usesPadding)?"yes":"no");
  
    printf("\nFrame histogram\n");
    for(i=0;i<15;i++)
        if (si.bitrateCount[i])
        {
            if (i==0)
                printf("Custom frames : ");
            else
                printf("%4d kbps : ",bitrate_tab[si.lsf][di.lay-1][i]);
            printf("%*d (%4.1f%%), size distr: [",max_len,si.bitrateCount[i],(double)si.bitrateCount[i]*100.0/(double)si.totFrameNum);

            if (si.unpadded[i].frameCount)
                printf("%*d x%4d B",max_len,si.unpadded[i].frameCount,si.unpadded[i].frameSize);

            if ((si.unpadded[i].frameCount)&&(si.padded[i].frameCount))
                printf(", ");

            if (si.padded[i].frameCount)
                printf("%*d x%4d B",max_len,si.padded[i].frameCount,si.padded[i].frameSize);

            printf("]\n");
        }

    printf("\n%d header errors.\n\n",si.nSyncError);

    if (di.encoder_string[0] != '\0') printf("Encoder string : %s\n\n",di.encoder_string);
}

/*
 * vbr detection
 * returns 0 if cbr encoding
 * otherwise (vbr) nonzero
 */
char detect_vbr_stream(streamInfo *si)
{
    int i,j=0;
    for (i=1; i<15; i++)
    {
        if (si->bitrateCount[i])
            j++;
    }
    return (j>1);
}

char *guesser(detectionInfo di)
{
    char *ret;
/*
 * Note: Xing and helix encoders count in their xing tag also the xing tag itself
 * as a valid frame, resulting in a +1 difference with actual audio frames amount.
 * You have the same behaviour with FhG encoders and their VBRI tag
 */
    if (di.encoder_string[0]=='G' || di.encoder_string[0]=='L')
    {
        if (di.encoder_string[0]=='G')
            ret = "Gogo";
        else
            ret = "Lame";
    }
    else
    {
        if (di.lay==3                  /* these enhancements are only found in layerIII */
            &&
            di.ancillaryData           /* both require more infos, hidden into ancillary data */
            &&
            (
            (di.id!=3 && (di.enhSignature[0]&PRO_SIGN_BYTE1)==PRO_SIGN_BYTE1 && (di.enhSignature[1]&PRO_SIGN_BYTE2)==PRO_SIGN_BYTE2)     /* case of mp3PRO */
            ||
            (di.id==3 && (di.freq==44100 || di.freq==48000) && (di.enhSignature[0]&SURR_SIGN_BYTE1)==SURR_SIGN_BYTE1 && (di.enhSignature[1]&SURR_SIGN_BYTE2)==SURR_SIGN_BYTE2) /* case of mp3Surround */
            )
           )
        {
            if (di.id!=3)
            {
                if (di.ofl)
                    ret = "Fraunhofer IIS mp3PRO encoder with OFL";
                else
                    ret = "Fraunhofer IIS mp3PRO encoder";
            }
            else
            {
                if (di.ofl)
                    ret = "Fraunhofer IIS mp3Surround 5.1 encoder with OFL";
                else
                    ret = "Fraunhofer IIS mp3Surround 5.1 encoder";
            }
        }
        else
        {
            if (di.vbr_tag==TAG_VBRITAG)
            {
                /* Presence of VBRI tag means we're dealing with a vbr stream produced by a fraunhofer encoder */
                if (di.ofl)
                    /* mp3sEncoder is the encoder for mp3surround but it can encode both mono and stereo streams as well */
                    ret = "FhG mp3sEncoder with OFL";
                else
                {
                    if (di.usesPadding)
                        //ret = "FhG (l3enc, fastenc or mp3enc)"; /* both l3enc and mp3enc cannot encode vbr streams */
                        ret = "FhG fastenc, mp3sEncoder or mp3 plugin";
                    else
                        ret = "FhG (ACM or producer pro)";
                }
            }
            else
            {
                /* "...the Xing encoder never uses short blocks." */
                if (di.blockCount[1] == 0)
                {  /* helix encoder (which is based on xing source code) DOES use short blocks! */
                    if (di.modeCount[1])   /* intensity stereo encoded frames */
                        ret = "Xing (very old)";
                    else
                    {
                        if (di.usesScfsi)
                            ret = "Xing (new)";
                        else
                            ret = "Xing (old)";
                    }
                }
                else
                {
                    if (di.usesScfsi)
                    {  /* gogo-no-coda (?) seems to be a japanese fork of lame using asm-optimizations in
                          order to provide max encoding speed; it adds NO lame tag and a custom "GOGO"
                          string in ancillary bits. Further, it adds a xing tag in vbr encodings only. */
                        if (di.usesScalefacScale)
                        {   /* the encoder may be lame as well as gogo as helix -- xing tag may help now */
                            if (di.vbr_tag==TAG_XINGTAG || di.vbr_tag==TAG_LAMECBRTAG)
                            {
                                if (di.lameTag)
                                    ret = "Lame";   /* ScaleFactor scaling is used in Lame since version 3.84 beta on */
                                else
                                {
                                    if (di.frameCountDiffers)
                                        ret = "Helix";
                                    else
                                        ret = "Gogo";
                                }
                            }
                            else
                                ret = "Helix";
                        }
                        else
                            ret = "Lame (old) or m3e"; /* Lame (up to) v3.83 beta didn't use it */
                    }
                    else
                    {
                        if (di.usesScalefacScale)
                        {
                            if (di.ofl)
                                ret = "FhG mp3sEncoder with OFL";
                            else
                            {
                                if (di.vbr_stream)
                                    ret = "FhG fastenc";
                                else
                                {
                                    if (di.usesPadding)
                                        ret = "FhG (l3enc, fastenc or mp3enc)";
                                    else
                                        ret = "FhG (ACM or producer pro)";
                                }
                            }
                        }
                        else
                            ret = "dist10 encoder or other encoder";
                    }
                }
            }
        }
    }
    return ret;
}

///////////////////////////////////////////////////////

void print_version(char verbose)
{
    printf ("%s-%u.%u",VER_PKG_NAME,VER_MAJ,VER_MIN);
    if (RELEASE_STABLE) printf(".%u final",VER_PATCH_LEV);
    if (RELEASE_ALPHA) printf (" alpha");
    if (RELEASE_BETA) printf (" beta");
    if (RELEASE_ALPHA || RELEASE_BETA) printf (" %u", VER_PATCH_LEV);
    if (verbose)
    {
#ifdef CODENAME
        printf(", codename \"%s\".",CODENAME);
#endif
        printf("\nExecutable built on %s.", __DATE__);
    }
    printf("\n");
}

void usage()
{
    print_version(0);
    printf("\
Show detailed informations about mpeg audio files and guess the encoder used.\n\
Usage:\t%s [ -h | -V | filename ]\n\
Options:\n\
  -h\tprint this help screen\n\
  -V\tprint version informations\n\
\n",VER_PKG_NAME);
}

///////////////////////////////////////////////////////
int main(int argc,char **argv)
{
    char buf[LARGEST_BUFFER], lame_tag_found=0;
    id3tag id3;
    unsigned char id3V2Maj, id3V2min;
    int i, option, head, id3v2size, apetagsize, apetagvers, apetagitems;
    off_t id3pos=0, pos=0;
    FILE *input_file;
    detectionInfo di;
    streamInfo si;
    VBRTAGDATA TagData;
#ifdef _WIN32
    struct _stati64 fileStat;
    #define stat _stati64
#else
    struct stat fileStat;
#endif /* _WIN32 */

    /* reset some values */
    di.usesScfsi=0;di.usesScalefacScale=0;di.usesPadding=0;di.ancillaryData=0;di.ofl=0;
    di.enhSignature[0]=255;di.enhSignature[1]=255;
    si.totFrameLen=0;si.totFrameNum=0;si.reservoirMax=0;si.nSyncError=0;si.freeformat=0;
    si.ancill_min=LARGEST_BUFFER;si.ancill_max=0;si.encDelay=-1;si.orig_samples=-1;
    si.min_global_gain[0]=255;si.min_global_gain[1]=255;si.max_global_gain[0]=0;si.max_global_gain[1]=0;

    opterr=0;
    while ((option = getopt(argc, argv, "hV")) != -1)
    {
#ifdef CODENAME
        char source[DATA_LEN];
#endif
        switch (option)
        {
            case 'h':
                usage();
                return 0;
            case 'V':
#ifdef CODENAME
                if (validate(argc,argv,optind,CMD_CODENAME_LENGTH,CODENAME))
                {
                    for (i=0; i<DATA_LEN; i++) source[i]=scrambled_data[i];
                    printf("%s",decode(source,DATA_LEN));
                }
                else
#endif
                    print_version(1);
                return 0;
            case '?':
                fprintf (stderr, "-%c: unknown option.\n", optopt);
                return -1;
            default:
                usage();
            return 0;
        }
    }

    if (optind < argc -1 )
    {  /* remaining args are input filenames */
        printf("WARNING: actually only the first filename will be used.\n");
    }
    if (optind == argc)
    { /* no filenames provided */
        usage();
        return 0;
    } /*if*/

    for (i=0; i<LAME_STRING_LENGTH; i++) di.encoder_string[i] = '\0';

    /* is `input_file' a regular file? */
    if (stat(argv[optind],&fileStat))
    {
        printf("Error getting file attributes for `%s' (does it exist?), exiting.\n",argv[optind]);
        return -1;
    }
    if (
    !S_ISREG(fileStat.st_mode)
#ifndef _WIN32
    &&!S_ISLNK(fileStat.st_mode)
#endif
    )
    {
        printf("The file `%s' is not a regular file, unable to proceed, exiting.\n",argv[optind]);
        return 1;
    }

    si.filesize = fileStat.st_size;

    input_file = fopen(argv[optind],"rb");
    if (input_file == NULL)
    {
        printf("Fatal: error opening `%s'.\nDouble-check the provided file name or ensure you have the proper access rights.\n", argv[optind]);
        return -1;
    }
    printf("Reading `%s'...\n",argv[optind]);

    /* check if this file has ID3 tag(s). */
    id3pos=checkid3v1(input_file,&id3);
    if (id3pos!=si.filesize)
    {
        printf("ID3tag v1.%c found.\n",(!id3.comment[28]&&id3.comment[29])?'1':'0');
        show_id3v1(id3);
    }

    apetagsize=checkapetagx(input_file,id3pos,&apetagvers,&apetagitems);
    if (apetagsize)
    {
        printf("APE tag v%c found at the end of the stream - %d items, %d bytes long.\n\n",(apetagvers==2000)?'2':'1',apetagitems,apetagsize);
        id3pos -= apetagsize;
    }

    id3v2size=checkid3v2(input_file,&id3V2Maj,&id3V2min);
    if (id3v2size>0)
    {
        printf("ID3tag v2.%01u.%01u found.\n",id3V2Maj,id3V2min);
        printf("ID3tag v2 is %d bytes long, skipping...\n\n",id3v2size);
    }
    else
    {
        if (id3v2size==-1)
        {
            printf("Errors detected in id3V2 tag header.\n\n");
            /* I will need the size of the whole tag (if present) into `id3v2size' */
            /* so in this case it has to be fixed */
            id3v2size=0;
        }
    }
    /* due to this `fseek', I will assure the file pointer points */
    /* at the beginning of the non-id3v2 stuff */
    fseeko(input_file,(off_t)id3v2size,SEEK_SET);

    // find first frame
    /*
     * lame encoder prior to version 3.94 beta (dec 15, 2003) used to set a wrong
     * bitrate index (1111b) in the vbr header for free format streams
     * I have to allow a more relaxed scan in order not to skip a valid vbr header
     */
    pos = (off_t)id3v2size-1;

    do
    {
        unsigned char c1,c2,c3,c4;
        off_t index;
        char seek_points[16]={0,4,4,8,100,104,104,108,4,8,8,12,104,108,108,112};/* offset of the contents of xing/lame tag depends upon tag flags */
        pos = resync_mp3(input_file,&head,pos+1,&i,1);
        if (pos!=-1 && ((head&HEADER_FIELD_LAYER)>>17)==1)
        {
            index=4+         /* 32 bit header */
                  (off_t)sideinfo_tab[((head&HEADER_FIELD_CHANNELS)>>6)/3][1-((head&HEADER_FIELD_LSF)>>19)]; /* side informations */
            fseeko(input_file,index,SEEK_CUR); /* seek to the xing vbr header */
            c1=getc(input_file);
            c2=getc(input_file);
            c3=getc(input_file);
            c4=getc(input_file);
            if (c1=='I' && c2=='n' && c3=='f' && c4=='o')
            {   /* tag found, now read the flags */
                c1=getc(input_file);
                c2=getc(input_file);
                c3=getc(input_file);
                c4=getc(input_file);
                if (!c1 && !c2 && !c3 && !(c4&0xf0)) /* vbr tag only uses lowest 4 bit */
                {
                    fseeko(input_file,(off_t)seek_points[c4],SEEK_CUR);
                    c1=getc(input_file);
                    c2=getc(input_file);
                    c3=getc(input_file);
                    c4=getc(input_file);
                    if (c1=='L' && c2=='A' && c3=='M' && c4=='E')
                        lame_tag_found=1;
                }
            }
        }
    }
    while (pos!=-1 && ((head&HEADER_FIELD_LAYER)>>17)==1 && ((head&HEADER_FIELD_BITRATE)>>12)==0x0f && !lame_tag_found);

    if (pos == -1)
    {
        printf("Cannot find valid mpeg header, scan failed.\n\n");
        fclose(input_file);
        return -1;
    }

    /* does the mpeg stream start right after ID3v2 ? */
    if (pos!=id3v2size)
        printf("Unexpected data at %d (length in bytes: %d).\n",id3v2size,(int)(pos-(off_t)id3v2size));

    fseeko(input_file,pos,SEEK_SET);
    i=(int)fread(&buf,1,LARGEST_BUFFER,input_file);
    if (GetInfoTag(&TagData,buf,i,pos,di.encoder_string))
    {
        show_info_tag(&TagData);
        di.vbr_tag=TagData.infoTag;
        di.lameTag=(TagData.lametag[0]!=0);
        fseeko(input_file,(off_t)(id3v2size+TagData.frameSize),SEEK_SET);
        pos = resync_mp3(input_file,&head,RESYNC_POS_CURRENT,&i,0);
        if (pos == -1)
        {
            printf("Cannot find more valid mpeg header, scan failed.\n\n");
            fclose(input_file);
            return -1;
        }
    }

    /* set the frame size in case of a free format bitstream,
       using the frame size detected (in any case) by GetInfoTag */
    if (freeformat(TagData.header))
    {
        int slot=((TagData.header&HEADER_FIELD_LAYER)==HEADER_FIELD_LAYER)?4:1;
        if (TagData.header&HEADER_FIELD_PADDING)
        {
            si.padded[0].frameSize=TagData.frameSize;
            si.unpadded[0].frameSize=TagData.frameSize-slot;
        }
        else
        {
            si.unpadded[0].frameSize=TagData.frameSize;
            si.padded[0].frameSize=TagData.frameSize+slot;
        }
    }

    printf("First frame found at %lld (0x%08X).\n",(long long)pos,(unsigned)pos);

    scan(input_file,&pos,&si,&di,id3pos,buf);

    if (fclose(input_file))
    {
        return -1;  /* check for errors on closing stream */
    }

    if (si.totFrameNum)
    {
        show_stream_info(si,di,&TagData);

        if (di.vbr_tag!=TAG_NOTAG)
            di.frameCountDiffers=(si.totFrameNum!=TagData.frames);
        else
            di.frameCountDiffers=0;

        di.vbr_stream=detect_vbr_stream(&si);

        if (di.lay==3)
            printf("Maybe this file is encoded by %s\n\n",guesser(di));
    }
    else
        printf("No valid audio data.\n\n");

    return 0;
}

